# -*- coding: utf-8 -*-
"""
File system related helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from os.path import exists, join, isdir
from shutil import rmtree
from fnmatch import fnmatch

# 3rd party imports
from six import string_types

# local imports
from . import log
from . import shell


def surround_paths_with_quotes(paths):
    """ Put quotes around all paths and join them with space in between. """
    if isinstance(paths, string_types):
        raise ValueError(
            "paths cannot be a string. "
            "Use array with one element instead."
        )
    return ' '.join('"' + path + '"' for path in paths)


def rm_glob(pattern, exclude_env=True, exclude_tox=True):
    """ Remove all files matching the given glob *pattern*. """
    log.info("Removing files matching {}", pattern)

    cmd = ['find . -name "{}"'.format(pattern)]

    if exclude_env:
        # Remove entries starting with ./env
        cmd.append("| sed '/^\.\/env/d'")

    if exclude_tox:
        # Remove entries starting with ./.tox
        cmd.append("| sed '/^\.\/\.tox/d'")

    matches = shell.run(' '.join(cmd), capture=True)

    for path in matches.stdout.splitlines():
        # might be a child of a dir deleted in an earlier iteration
        if not exists(path):
            continue

        if not isdir(path):
            log.info('  ^91[file] ^90{}'.format(path))
            os.remove(path)
        else:
            log.info('  ^91[dir]  ^90{}'.format(path))
            rmtree(path)


def filtered_walk(path, exclude):
    if not isdir(path):
        raise ValueError("Cannot walk files, only directories")

    files = os.listdir(path)
    for name in files:
        file_path = join(path, name)

        if is_excluded(name, file_path, exclude):
            continue

        yield name, file_path

        if isdir(file_path):
            for n, p in filtered_walk(file_path, exclude):
                yield n, p


def is_excluded(name, path, excluded):
    matches = lambda pattern: fnmatch(name, pattern) or fnmatch(path, pattern)
    return next((True for x in excluded if matches(x)), False)



# -*- coding: utf-8 -*-
"""
Testing commands
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
import os.path

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import shell
from . import cli


BUILD_DIR = conf.get_path('BUILD_DIR', '.build')

PYTEST_CFG_PATH = conf.get_path('PYTEST_CFG_PATH', 'ops/tools/pytest.ini')
TEST_TYPES = conf.get('TEST_TYPES', {})

COVERAGE_OUT_PATH = os.path.join(BUILD_DIR, 'coverage')
COVERAGE_CFG_PATH = conf.get_path('COVERAGE_CFG_PATH', 'ops/tools/coverage.ini')

DJANGO_SETTINGS = conf.get('DJANGO_SETTINGS', None)
DJANGO_TEST_SETTINGS = conf.get('DJANGO_TEST_SETTINGS', None)


@cli.command()
@click.option('--no-sugar', is_flag=True)
@click.option('--type', type=str, default='default')
@click.option('--verbose', count=True)
@click.option('--junit', is_flag=True)
@click.option('--no-locals', is_flag=True)
@click.option('--no-coverage', is_flag=True)
@click.option('--allow-empty', is_flag=True)
@click.option('--plugins', type=str, default='')
# def test(**opts):
def test(
        no_sugar, type, verbose, junit, no_locals,
        no_coverage, plugins, allow_empty
):
    """ Run all tests against the current python version. """
    SRC_DIR = conf.get_path('SRC_DIR')
    SRC_PATH = conf.get_path('SRC_PATH')
    plugins = plugins.split(',')
    args = []

    if not no_coverage:
        args += [
            '--durations=3',
            '--cov-config={}'.format(COVERAGE_CFG_PATH),
            '--cov={}'.format(SRC_PATH),
            '--cov-report=term:skip-covered',
            '--cov-report=html:{}'.format(COVERAGE_OUT_PATH),
        ]

    if junit:
        args += ['--junitxml={}/test-results.xml'.format('.build')]

    if '-django' not in plugins:
        if DJANGO_TEST_SETTINGS is not None:
            args += ['--ds {}'.format(DJANGO_TEST_SETTINGS)]
        elif DJANGO_SETTINGS is not None:
            args += ['--ds {}'.format(DJANGO_SETTINGS)]

    if no_sugar:
        args += ['-p no:sugar']

    if verbose >= 1:
        args += ['-v']
    if verbose >= 2:
        args += ['--full-trace']

    if not no_locals:
        args += ['-l']

    if plugins:
        for plug_name in plugins:
            if not plug_name.strip():
                continue

            if plug_name.startswith('-'):
                args += ['-p no:{}'.format(plug_name[1:])]
            else:
                args += ['-p {}'.format(plug_name)]

    test_config = {'paths': SRC_PATH}
    if type is not None:
        test_config = TEST_TYPES.get(type)
        mark = test_config.get('mark')

        if mark:
            args += ['-m "{}"'.format(mark)]

    test_paths = test_config['paths'] or []
    test_paths = [conf.proj_path(p) for p in test_paths]
    result = shell.run(
        'pytest -c {conf} {args} {paths}'.format(
            conf=PYTEST_CFG_PATH,
            args=' '.join(args),
            paths = fs.surround_paths_with_quotes(test_paths)
        ),
        env={'PYTHONPATH': SRC_DIR},
        exit_on_error=False
    )

    if result.failed and not (allow_empty is True and result.return_code == 5):
        sys.exit(result.return_code)



# -*- coding: utf-8 -*-
""" Frontend related commands.

All frontend commands should be implemented through
``src/frontend/package.json``. This is just a proxy to allow easily running
them from the root project directory.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from os.path import exists, join

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import log
from peltak.core import shell
from . import cli


FRONTEND_PATH = conf.get_path('FRONTEND_PATH', None)
FRONTEND_CMDS = conf.get('FRONTEND_CMDS', {
    'build': 'npm run build',
    'start': 'npm start',
    'watch': 'npm run watch',
    'test': 'npm test',
    'lint': 'npm lint',
    'check': 'npm run lint && npm run test'
})


def _fe_cmd(cmd):
    if FRONTEND_PATH is not None:
        with conf.within_proj_dir(FRONTEND_PATH):
            shell.run(cmd)
    else:
        log.err("No FRONTEND_PATH defined in the config")


@cli.command()
@click.argument('cmd')
def fe(cmd):
    """ Run a predefined frontend command.

    The commands can be defined through the FRONTEND_CMDS project configuration
    value. It should be a dict of commands mapped to the actual scripts ran
    inside the FRONTEND_DIR.
    """
    if cmd in FRONTEND_CMDS:
        _fe_cmd(FRONTEND_CMDS[cmd])
    else:
        log.err("No {} in FRONTEND_CMDS".format(cmd))


@cli.command('init-fe')
@click.argument('cmd')
@click.option('--no-recreate', is_flag=True)
def init_fe(no_recreate=False):
    """ Initialize frontend. """
    initialized = exists(join(FRONTEND_PATH, 'node_modules'))

    if not (initialized and no_recreate):
        _fe_cmd('npm install')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import click

import peltak


@click.group()
@click.version_option(version=peltak.__version__, message='%(version)s')
def cli():
    """

    To get help for a specific command:

        \033[1mpeltak <command> --help\033[0m

    Examples:

        \033[1mpeltak lint --help\033[0m

        \033[1mpeltak version bump --help\033[0m

        \033[1mpeltak release upload --help\033[0m

    """
    pass



# -*- coding: utf-8 -*-
"""
Code linting commands.
"""
from __future__ import absolute_import, unicode_literals

# 3rd party imports
from six import string_types

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import log
from peltak.core import shell
from . import cli


PYLINT_CFG_PATH = conf.get_path('PYLINT_CFG_PATH', 'ops/tools/pylint.ini')
PEP8_CFG_PATH = conf.get_path('PEP8_CFG_PATH', 'ops/tools/pep8.ini')
LINT_PATHS = conf.get('LINT_PATHS', [])


def _lint_files(paths):
    """ Run static analysis on the given files.

    :param paths:   Iterable with each item being path that should be linted..
    """
    if isinstance(paths, string_types):
        raise ValueError("paths must be an array of strings")

    log.info("Linting")
    for path in paths:
        print("--   {}".format(path))

    paths = fs.surround_paths_with_quotes(paths)

    log.info("Checking PEP8 compatibility")
    pep8_cmd = 'pep8 --config {} {{}}'.format(PEP8_CFG_PATH)
    pep8_ret = shell.run(pep8_cmd.format(paths)).return_code

    log.info("Running linter")
    pylint_cmd = 'pylint --rcfile {} {{}}'.format(PYLINT_CFG_PATH)
    pylint_ret = shell.run(pylint_cmd.format(paths)).return_code

    if pep8_ret != 0:
        print("pep8 failed with return code {}".format(pep8_ret))

    if pylint_ret:
        print("pylint failed with return code {}".format(pylint_ret))

    return pep8_ret == pylint_ret == 0


@cli.command()
def lint():
    """ Run pep8 and pylint on all project files. """
    if not _lint_files([conf.proj_path(p) for p in LINT_PATHS]):
        exit(1)



# -*- coding: utf-8 -*-
"""
Helper commands for releasing to pypi.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import sys
from os.path import join
from warnings import warn

# 3rd party modules
import click

# local imports
from peltak.core import shell
from peltak.core import conf
from peltak.core import git
from peltak.core import log
from peltak.core import versioning
from . import cli


VERSION_FILE = conf.get_path('VERSION_FILE', 'VERSION')


@cli.group('release')
def rel():
    """ Release related commands. """
    pass


@rel.command('make')
@click.argument(
    'component',
    type=click.Choice(['major','minor', 'patch']),
    required=False,
    default='patch'
)
@click.option('--exact', type=str)
def make_release(component, exact):
    """ Release a new version of the project.

    This will bump the version number (patch component by default) + add and tag
    a commit with that change. Finally it will upload the package to pypi.

    1. Bump version.
    2. Create and checkout release/* branch
    3. Create commit with bumped version.
    """
    with conf.within_proj_dir(quiet=True):
        out = shell.run('git status --porcelain', capture=True).stdout
        has_changes = any(
            not l.startswith('??') for l in out.split(os.linesep) if l.strip()
        )

    if has_changes:
        log.info("Cannot release: there are uncommitted changes")
        exit(1)

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Bumping package version")
    log.info("  old version: ^35{}".format(old_ver))
    log.info("  new version: ^35{}".format(new_ver))

    with conf.within_proj_dir(quiet=True):
        branch = 'release/' + new_ver

        log.info("Checking out new branch ^35{}", branch)
        shell.run('git checkout -b ' + branch)

        log.info("Creating commit for the release")

        shell.run('git add {ver_file} && git commit -m "{msg}"'.format(
            ver_file=VERSION_FILE,
            msg="Release: v{}".format(new_ver)
        ))


@rel.command('tag')
def tag_release():
    """ Create a new release tag for the current version. """
    release_ver = versioning.current()
    author = git.commit_author()

    with conf.within_proj_dir(quiet=False):
        log.info("Creating tag that marks the release")
        cmd = (
            'git -c "user.name={0.name}" -c "user.email={0.email}" '
            'tag -a "v{1}" -m "Mark v{1} release"'
        ).format(
            author,
            release_ver
        )
        shell.run(cmd)


@rel.command()
@click.argument('target')
def upload(target):
    """ Release to a given pypi server ('local' by default). """
    log.info("Uploading to pypi server ^33{}".format(target))
    with conf.within_proj_dir(quiet=False):
        shell.run('python setup.py sdist register -r "{}"'.format(target))
        shell.run('python setup.py sdist upload -r "{}"'.format(target))


@rel.command('gen-pypirc')
@click.argument('username', required=False)
@click.argument('password', required=False)
def gen_pypirc(username=None, password=None):
    """ Generate .pypirc config with the given credentials. """
    path = join(conf.getenv('HOME'), '.pypirc')
    username = username or conf.getenv('PYPI_USER', None)
    password = password or conf.getenv('PYPI_PASS', None)

    if username is None or password is None:
        log.err("You must provide $PYPI_USER and $PYPI_PASS")
        sys.exit(1)

    log.info("Generating .pypirc config ^94{}".format(path))

    with open(path, 'w') as fp:
        fp.write('\n'.join((
            '[distutils]',
            'index-servers = ',
            '    pypi',
            '',
            '[pypi]',
            'repository: https://upload.pypi.org/legacy/',
            'username: {}'.format(username),
            'password: {}'.format(password),
            '',
        )))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# 3rd party modules
import click

# local imports
from peltak.core import log, conf, versioning
from . import cli


@cli.group('version')
def ver():
    """ Versioning related commands. """
    pass


@ver.command('show')
@click.option('--porcelain', is_flag=True)
def version(porcelain):
    """ Return current project version. """
    current = versioning.current()

    if porcelain:
        print(current)
    else:
        log.info("Version: ^35{}".format(current))


@ver.command('bump')
@click.argument(
    'component',
    type=click.Choice(['major','minor', 'patch']),
    required=False,
    default='patch'
)
@click.option('--exact', type=str)
def bump_version(component='patch', exact=None):
    """ Bump current project version without committing anything.

    No tags are created either.
    """

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Bumping package version")
    log.info("  old version: ^35{}".format(old_ver))
    log.info("  new version: ^35{}".format(new_ver))



# -*- coding: utf-8 -*-
"""
Various commands that do not fit another category but there's not enough of them
to justify a separate module.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os

# local imports
from peltak.core import conf
from peltak.core import fs
from . import cli


CLEAN_PATTERNS = conf.get('CLEAN_PATTERNS', [
    '__pycache__',
    '*.py[cod]',
    '.swp',
])


@cli.command()
def clean():
    """ Remove temporary files like python cache, swap files, etc. """
    cwd = os.getcwd()

    os.chdir(conf.proj_path('.'))

    for pattern in CLEAN_PATTERNS:
        fs.rm_glob(pattern)

    os.chdir(cwd)
